"""Preconfigured converters for the stdlib json."""
from cattrs.preconf.json import configure_converter, JsonConverter, make_converter

__all__ = ["configure_converter", "JsonConverter", "make_converter"]
