from cattrs.errors import (
    BaseValidationError,
    ClassValidationError,
    ForbiddenExtraKeysError,
    IterableValidationError,
    StructureHandlerNotFoundError,
)

__all__ = [
    "BaseValidationError",
    "ClassValidationError",
    "ForbiddenExtraKeysError",
    "IterableValidationError",
    "StructureHandlerNotFoundError",
]
